# Generates files for various benchmarks.
# For each benchmark multiple identical files are generated to
# make it possible to measure performance for different styles
# of encoding timestamps.

import sys

def pred(n,x):
    return (n*"pred(" + x + n*")")

# pred^n(t) <= t' => \/_{k<=n} pred^k(t) = t'
def pred_dicho(n):
    s = f"lemma [any] pred_dicho_{n} (t,t':timestamp) :\n  "
    s += pred(n,"t") + "<= t' => ("
    for k in range(1,n+1):
        s += pred(k,"t") + "= t' ||"
    s += "t<=t').\n"
    s += "Proof. try auto. Abort.\n"
    return s

# Same as pred_dicho but with missing first equality.
def pred_dicho_missfirst(n):
    s = f"lemma [any] pred_dicho_missfirst_{n} (t,t':timestamp) :\n  "
    s += pred(n,"t") + "<= t' => ("
    for k in range(2,n+1):
        s += pred(k,"t") + "= t' ||"
    s += "t<=t').\n"
    s += "Proof. try auto. Abort.\n"
    return s

# Reasoning using antisymmetry, transitivity and congruence close.
def trans_congr(n):
    variables = [ f"t_{i}" for i in range(0,n+1) ]
    s = f"lemma [any] trans_congr_{n} :\n"
    for t in variables:
        s += f"  forall ({t}:timestamp),\n"
    s += "  (\n"
    for k in range(0,n):
        s += f"   pred(t_{k}) <= t_{k+1} && t_{k+1} <= pred(t_{k}) &&\n"
    s += "  true) =>\n"
    s += "  " + pred(n,"t_0") + f" = t_{n}.\n"
    s += "Proof. try auto. Abort.\n"
    return s
    
# Same as trans_congr missing the first inequality
def trans_congr_missfirst(n):
    variables = [ f"t_{i}" for i in range(0,n+1) ]
    s = f"lemma [any] trans_congr_{n} :\n"
    for t in variables:
        s += f"  forall ({t}:timestamp),\n"
    s += "  (\n"
    for k in range(1,n):
        s += f"   pred(t_{k}) <= t_{k+1} && t_{k+1} <= pred(t_{k}) &&\n"
    s += "  true) =>\n"
    s += "  " + pred(n,"t_0") + f" = t_{n}.\n"
    s += "Proof. try auto. Abort.\n"
    return s
 
# (t<=t' /\ init<=pred^n(t)) => pred^n(t)<=pred^n(t')
def pred_compat(n):
    s = f"lemma [any] pred_compat_{n} (x,y:timestamp) :\n"
    px = pred(n,"x")
    py = pred(n,"y")
    s += f"  init <= {px} =>"
    s += f"  x <= y =>"
    s += f"  {px} <= {py}.\n"
    s += "Proof. try auto. Abort.\n"
    return s

benchmarks = {
  "pred_dicho": pred_dicho,
  "pred_dicho_missfirst": pred_dicho_missfirst,
  "trans_congr": trans_congr,
  "trans_congr_missfirst": trans_congr_missfirst,
  "pred_compat": pred_compat
}

if __name__=="__main__":
    sizes = [500]
    if len(sys.argv) > 1:
        sizes = [int(sys.argv[i]) for i in range(1,len(sys.argv))]
    for size in sizes:
        for style in ["Abs","Nat"]:
            for name in benchmarks:
                f = open(f"{name}_{size}_{style}.sp","w")
                f.write(benchmarks[name](size))
                f.close()
